/*********************************************
 * Some Python -> C bindings for PCI Utilities
 * 2006 - Joseph Jezak
 *********************************************/

#include <Python.h>
#include <pci/pci.h>

typedef struct {
	PyObject_HEAD
	struct pci_access *pacc;
}pci_PCIObject;
	
/* Locally raised exceptions */
static PyObject *PCIError;
#define onError(message) \
	{ PyErr_SetString(ErrorObject, message); return NULL; }

/* Return a PCI Device corresponding to the bus, dev and func passed in */
static struct pci_dev * find_pci_dev(struct pci_access *pacc, u8 bus, u8 dev, u8 func) {
	struct pci_dev *d;

	/* Fill each PCI device, then compare to see if we found it */
	for(d=pacc->devices; d; d=d->next) {
		pci_fill_info(d, PCI_FILL_IDENT | PCI_FILL_BASES);
		if(d->bus == bus && d->dev == dev && d->func == func)
			return d;
	}
	return NULL;
}

/* Get the Vendor ID for a device */
static u16 pci_vendor_id(struct pci_dev *d) {
	if (d != NULL)	
		return d->vendor_id;
	else
		return 0;
}

/* Get the Device ID for a device */
static u16 pci_device_id(struct pci_dev *d) {
	if (d != NULL)
		return d->device_id;
	else
		return 0;
}

/******************
 * Python Bindings
 ******************/

/* Allocate on init */
static PyObject * PCI_new(PyTypeObject *type, PyObject *args, PyObject *kwds) {
	pci_PCIObject *self;	

	self = (pci_PCIObject *)type->tp_alloc(type, 0);

	if(self != NULL) {
		/* Setup the PCI device */
		self->pacc = pci_alloc();
		pci_init(self->pacc);
	
		/* Scan the PCI bus, fills the PCI access struct */
		pci_scan_bus(self->pacc);
	}
	return (PyObject *)self;
}

/* Clean up alloc'd stuff */
static void PCI_dealloc(pci_PCIObject *self) {
	if(self->pacc != NULL)
		pci_cleanup(self->pacc);
	self->ob_type->tp_free((PyObject *)self);
}

/* *****************************************
 * Return the list of all of the PCI devices 
 * In the form of tuples, [0,0,0]
 * Takes no input
 * *****************************************/
static PyObject *
list_pci_devs(pci_PCIObject *self) {
	struct pci_dev *d;
	PyObject *dev_list;
	int pci_count = 0;
	
	/* Create a new list item */
	dev_list = PyList_New(0);
	if(dev_list == NULL)
		return NULL;

	/* Loop over each device and create bus,dev,func tuples */
	for(d=self->pacc->devices; d; d=d->next) 
		/* FIXME Add Error Checking! */
		PyList_Append(dev_list, Py_BuildValue("iii", d->bus, d->dev, d->func));

	return dev_list;
}

/* ******************************************
 * Get the PCI Config space data for a device 
 * Returns a string of the config space
 * ******************************************/
#define MAX_BUF		128
#define PCI_CONFIG_SZ	64
static PyObject *
get_config_info(pci_PCIObject *self, PyObject *args) {
	struct pci_dev *d;
	u8 bus, dev, func;
	u8 buf[MAX_BUF];
	
	int buf_size = PCI_CONFIG_SZ; 
	PyObject *py_buf;
	void *buf_ptr;

	if (!PyArg_ParseTuple(args, "bbb", &bus, &dev, &func))
		return NULL;

	/* Find the PCI Device */
	d = find_pci_dev(self->pacc, bus, dev, func);
	if(d == NULL)
		return NULL;
	
	/* Read the config space */
	if(!pci_read_block(d, 0, buf, PCI_CONFIG_SZ))
		return NULL;
	
	/* If this is a cardbus card, we need to get 64 more bytes */
	if((buf[PCI_HEADER_TYPE] & 0x7F) == PCI_HEADER_TYPE_CARDBUS){
		if(!pci_read_block(d, PCI_CONFIG_SZ, buf + PCI_CONFIG_SZ, PCI_CONFIG_SZ))
			return NULL;
		buf_size = MAX_BUF;
	}

	py_buf = PyBuffer_New(buf_size);
	PyObject_AsWriteBuffer(py_buf, &buf_ptr, &buf_size);	
	memcpy(buf_ptr, buf, buf_size);

	return py_buf;
}
#undef PCI_CONFIG_SZ
#undef MAX_BUF

/* Get Vendor ID for a device */
static PyObject *
get_vendor_id(pci_PCIObject *self, PyObject *args) {
	struct pci_dev *d;
	u8 bus, dev, func;
	int size = 0;
	u8 * buf;

	if (!PyArg_ParseTuple(args, "bbb", &bus, &dev, &func))
		return NULL;
	
	/* Find the PCI Device */
	d = find_pci_dev(self->pacc, bus, dev, func);
	if(d == NULL)
		return NULL;

	return Py_BuildValue("i", pci_vendor_id(d));
}

/* Get Device ID for a device */
static PyObject *
get_device_id(pci_PCIObject *self, PyObject *args) {
	struct pci_dev *d;
	u8 bus, dev, func;
	int size = 0;
	u8 * buf;

	if (!PyArg_ParseTuple(args, "bbb", &bus, &dev, &func))
		return NULL;
	
	/* Find the PCI Device */
	d = find_pci_dev(self->pacc, bus, dev, func);

	return Py_BuildValue("i", pci_device_id(d));
}


/* Exported Methods */
static struct PyMethodDef PCI_methods[] = {
	{ "get_config_info", (PyCFunction)get_config_info, METH_VARARGS, "Gets PCI config info for a device"},
	{ "get_device_id", (PyCFunction)get_device_id, METH_VARARGS, "Gets the Device ID of a device"},
	{ "get_vendor_id", (PyCFunction)get_vendor_id, METH_VARARGS, "Gets the Vendor ID of a device"},
	{ "list_pci_devs", (PyCFunction)list_pci_devs, METH_NOARGS, "Lists available PCI devices"},
	{ NULL, NULL, 0, NULL},
};

/* Exported Module Methods */
static struct PyMethodDef module_methods[] = {
	{NULL, NULL, 0, NULL}, 
};

static PyTypeObject pci_PCIType = {
	PyObject_HEAD_INIT(NULL)
	0,				/* ob_size */
	"pci.PCI",			/* tp_name */	
	sizeof(pci_PCIObject),		/* tp_basic_size */
	0,				/* tp_item_size */	
	(destructor)PCI_dealloc,	/* tp_dealloc */
	0,				/* tp_print */ 
	0,				/* tp_getattr */
	0,				/* tp_setattr */
	0,				/* tp_compare */
	0,				/* tp_repr */
	0,				/* tp_as_number */
	0,				/* tp_as_sequence */
	0,				/* tp_as_mapping */
	0,				/* tp_hash */
	0,				/* tp_call */
	0,				/* tp_str */
	0,				/* tp_getattro */
	0,				/* tp_set_attro */
	0,				/* tp_as_buffer */
	Py_TPFLAGS_DEFAULT,		/* tp_flags */
	"PCI Objects",			/* tp_doc */
	0,				/* tp_traverse */
	0,				/* tp_clear */
	0,				/* tp_richcompare */
	0,				/* tp_weaklistoffset */
	0,				/* tp_iter */
	0,				/* tp_iternext */
	PCI_methods,			/* tp_methods */
	0,
	0,
	0,
	0, 
	0,
	0,
	0,
	0,
	0,
	PCI_new,			/* tp_new */	

};

/* Init Module */
PyMODINIT_FUNC
initpci(void) {
	PyObject *m;

	/* Create the module and add functions */
	if (PyType_Ready(&pci_PCIType) < 0)
		return;

	m = Py_InitModule3("pci", module_methods, "Allows access to PCI Data");
	Py_INCREF(&pci_PCIType);
	PyModule_AddObject(m, "PCI", (PyObject *)&pci_PCIType);

	PCIError = PyErr_NewException("pci.error", NULL, NULL);
	Py_INCREF(PCIError);
	PyModule_AddObject(m, "error", PCIError);
}
